import numpy as np

import torch
from torch.nn.utils.rnn import pad_sequence

        
class OneHotEncode(torch.nn.Module):
    """OneHot Encode an integer"""

    def __init__(self, n_vocab):
        super().__init__()
        self.n_vocab = n_vocab

    def forward(self, x):
        one_hot = np.zeros((np.multiply(*x.shape), self.n_vocab), dtype=np.float32)
        one_hot[np.arange(one_hot.shape[0]), x.flatten()] = 1.
        one_hot = one_hot.reshape((*x.shape, self.n_vocab))
        one_hot = torch.FloatTensor(one_hot)
        return one_hot


class TruncateSequence(torch.nn.Module):
    """Truncate sequence data"""
    def __init__(self, n_truncate):
        super().__init__()
        self.n_truncate = n_truncate

    def forward(self, x):
        if x.dim() == 3:
            Xprior = x[:, :-self.n_truncate, :]
            Xtrunc = x[:, -self.n_truncate:, :]
        elif x.dim() == 2:
            Xprior = x[:, :-self.n_truncate]
            Xtrunc = x[:, -self.n_truncate:]
        else:
            raise Exception("Input must be in dimension 2 or 3")
        return (Xprior, Xtrunc)
        

class PadSequence(torch.nn.Module):
    """Pad sequences to the longest sequence"""
    def __init__(self, pad_id=0):
        super().__init__()
        self.pad_id = pad_id
    
    def forward(self, x):
        x = [torch.LongTensor(seq) for seq in x]
        return pad_sequence(x, batch_first=True, padding_value=self.pad_id)




import torch
import pandas as pd
from torch.utils.data import Dataset, DataLoader


class CharRNNDataset(Dataset):
    """
    CharRNN Dataset for PyTorch
    
    Process a csv file with:
    - a column of string
    - a column of label for the string
    For example:
        __________________________
        text     | label
        --------------------------
        C1D2F3G4 | pop
        D1G2F1C4 | jazz
        ...      | ...
        __________________________

    == Arguments ==
    csv_path: string
        the dataset

    chars: string or list[char]
        list of unique characters in the dataset. If None, would be derived from dataset (may be computatioally expensive)

    classes: list
        list of prediction classes. If None, would be derived from dataset

    text_col: str
        column header in the csv_path for the text column

    label_col: str
        column header in the csv_path for the label column        

    pad: char
        a character to be used as the padding token

    max_len: str
        exclude data with more than max_len characters. If None, all data is used
    """    
    def __init__(self, csv_path, chars=None, classes=None, ascending=False, text_col='text', label_col="label", pad="-", max_len=None):
        df = pd.read_csv(csv_path)
        assert text_col in df.columns
        assert label_col in df.columns
        df = df.reindex(df[text_col].str.len().sort_values(ascending=ascending).index)        

        self.pad = pad
        
        if classes is None:
            classes = sorted(df[label_col].unique().tolist())
        self.classes = classes
        self.n_classes = len(classes)        
        self.class2idx = dict(zip(classes, range(len(classes))))
        
        if chars is None:
            chars = sorted(set("".join(df[text_col].to_list())))
        if chars[0] != pad:
            chars = [pad] + chars
        self.chars = chars
        self.n_chars = len(chars)
        self.char2idx = dict(zip(chars, range(len(chars))))
        
        self.X = [self.encode_string(row) for row in df[text_col]]
        self.y = [self.encode_class(row) for row in df[label_col]]
        
        if max_len is not None:
            if ascending:
                for i, data in enumerate(self.X):
                    if len(data) > max_len:
                        break
                self.X = self.X[:i]
                self.y = self.y[:i]
            else:
                for i, data in enumerate(self.X):
                    if len(data) < max_len:
                        break
                self.X = self.X[i:]
                self.y = self.y[i:]

    def __getitem__(self, index):        
        return self.X[index], self.y[index]

    def __len__(self):
        return len(self.X)
    
    def encode_string(self, text):
        return [self.char2idx[char] for char in text]
    
    def encode_class(self, class_str):
        return self.class2idx[class_str]
    
    def decode_class(self, class_id):
        return self.classes[class_id]
    

class CharRNNDataloader(DataLoader):
    """
    CharRNN Dataloader for PyTorch

    == Arguments ==
    dataset: CharRNNDataset
        the CharRNN dataset object

    batch_size: int
        how many item to sample in each batch

    shuffle: bool
        if True, random sampling is used, sequential sampling is used otherwise.
        It will be neglected when class_sample_weight is provided.

    batch_tranform: transforms.Compose
        list of sequential transformation applied at batch level

    sampler: Sampler or Iterable
        defines the strategy to draw samples from the dataset. Can be any ``Iterable`` with ``__len__`` implemented.
        If specified, :attr:`shuffle` must not be specified.
    
    drop_last: bool
        If True, drop the last incomplete batch

    num_workers: int
        how many subprocesses to use for data loading.
        0 means that the data will be loaded in the main process.
    """        
    def __init__(self, dataset, batch_size=1, shuffle=False, batch_transform=None, sampler=None, drop_last=True, num_workers=0):
        self.transform = batch_transform
        super().__init__(dataset, batch_size, shuffle, sampler=sampler, collate_fn=self.collate, num_workers=num_workers, drop_last=drop_last)

    def collate(self, batch):
        Xb, yb = zip(*batch)
        
        yb = torch.LongTensor(yb)
        
        if self.transform is not None:
            Xb = self.transform(Xb)                
        return Xb, yb



from torch import nn

from jcopdl.layers._base import BaseBlock


class LinearBlock(BaseBlock):
    """
    available activation {relu, lrelu, sigmoid, tanh, elu, selu, lsoftmax, None}
    """
    def __init__(self, n_input, n_output, activation="relu", batchnorm=False, dropout=0):
        super().__init__()
        self.block = nn.Sequential()
        self.block.add_module("linear", nn.Linear(n_input, n_output))

        if batchnorm:
            self.append_batchnorm(n_output)

        self.append_activation(activation)

        if dropout > 0:
            self.append_dropout(dropout)

    def forward(self, x):
        return self.block(x)


class ConvBlock(BaseBlock):
    """
    available activation {relu, lrelu, sigmoid, tanh, elu, selu, lsoftmax, None}
    available pool_type {maxpool, avgpool, None}
    """
    def __init__(self, in_channel, out_channel, kernel=3, stride=1, pad=1, bias=True, activation="relu",
                 batchnorm=False, dropout=0, pool=None, pool_kernel=2, pool_stride=2):
        super().__init__()
        self.block = nn.Sequential()
        self.block.add_module("conv2d", nn.Conv2d(in_channel, out_channel, kernel, stride, pad, bias=bias))

        if batchnorm:
            self.append_batchnorm2d(out_channel)

        self.append_activation(activation)

        if dropout > 0:
            self.append_dropout2d(dropout)

        self.append_pooling2d(pool, pool_kernel, pool_stride)

    def forward(self, x):
        return self.block(x)


class TConvBlock(BaseBlock):
    def __init__(self, in_channel, out_channel, kernel=3, stride=1, pad=1, bias=True, activation="relu",
                 batchnorm=False, dropout=0, pool=None, pool_kernel=2, pool_stride=2):
        super().__init__()
        self.block = nn.Sequential()
        self.block.add_module("tconv2d", nn.ConvTranspose2d(in_channel, out_channel, kernel, stride, pad, bias=bias))

        if batchnorm:
            self.append_batchnorm2d(out_channel)

        self.append_activation(activation)

        if dropout > 0:
            self.append_dropout2d(dropout)

        self.append_pooling2d(pool, pool_kernel, pool_stride)

    def forward(self, x):
        return self.block(x)


class VGGBlock(BaseBlock):
    def __init__(self, in_channel, out_channel, n_repeat=2, kernel=3, stride=1, pad=1, bias=True, activation="relu",
                 batchnorm=False, dropout=0, pool="maxpool", pool_kernel=2, pool_stride=2):
        super().__init__()
        self.block = nn.Sequential()
        self.block.add_module("conv_block0",
                              ConvBlock(in_channel, out_channel, kernel, stride, pad, bias, activation, batchnorm, dropout))

        for i in range(1, n_repeat):
            if i == (n_repeat - 1):
                self.block.add_module(f"conv_block{i}",
                                      ConvBlock(out_channel, out_channel, kernel, stride, pad, bias, activation, batchnorm,
                                                dropout, pool, pool_kernel, pool_stride))
            else:
                self.block.add_module(f"conv_block{i}",
                                      ConvBlock(out_channel, out_channel, kernel, stride, pad, bias, activation, batchnorm,
                                                dropout))

    def forward(self, x):
        return self.block(x)



import numpy as np
from torch.utils.data.sampler import WeightedRandomSampler


class BalancedRandomSampler(WeightedRandomSampler):
    """
    Automatically compute balanced weight for PyTorch WeightedRandomSampler.
    The 'balanced' heuristic is explained ini
    King, G. & Zeng, L. 2001. Logistic Regression in Rare Events Data
    https://dash.harvard.edu/bitstream/handle/1/4125045/relogit?sequence=2


    == Arguments ==
    y: np.array
        label encoded data

    replacement: bool
        If ``True``, samples are drawn with replacement.
        If not, they are drawn without replacement, which means that when a sample index
        is drawn for a row, it cannot be drawn again for that row.
    """
    def __init__(self, y, replacement: bool = True):
        n_classes = len(set(y))
        recip_freq = len(y) / (n_classes * np.bincount(y).astype(np.float64))
        weights = recip_freq[y]
        super().__init__(weights, num_samples=len(weights), replacement=replacement)



import torch
import numpy as np
from scipy.sparse import coo_matrix


class MiniBatchBinaryConfusionMatrix():
    def __init__(self):
        self.y_true = []
        self.y_pred = []
    
    def add_batch(self, batch_preds, batch_targets):
        """
        batch_preds: (N, F)
        batch_targets: (N,)
        """
        assert batch_preds.ndim == 2
        self.y_pred.append(batch_preds)
        self.y_true.append(batch_targets)
        
    def compute(self):
        """
         |0_|1_|
        0|__|__|
        1|__|__|
        """
        yp = torch.cat(self.y_pred).argmax(1)
        yt = torch.cat(self.y_true)
        
        fn = ((yt == 1) & (yp == 0)).sum().item()
        tp = ((yt == 1) & (yp == 1)).sum().item()
        fp = ((yt == 0) & (yp == 1)).sum().item()
        tn = ((yt == 0) & (yp == 0)).sum().item()       
        return np.array([[tn, fp], [fn, tp]])
    

class MiniBatchConfusionMatrix():
    def __init__(self):
        self.y_true = []
        self.y_pred = []
    
    def add_batch(self, batch_preds, batch_targets):
        """
        batch_preds: (N, F)
        batch_targets: (N,)
        """
        assert batch_preds.ndim == 2
        self.y_pred.append(batch_preds)
        self.y_true.append(batch_targets)
        
    def compute(self):
        """
         |0_|1_|2_|
        0|__|__|__|
        1|__|__|__|
        2|__|__|__|
        """
        yp = torch.cat(self.y_pred).argmax(1).numpy()
        yt = torch.cat(self.y_true).numpy()

        sample_weight = np.ones(yt.shape[0], dtype=np.int64)
        cm = coo_matrix((sample_weight, (yt, yp)), dtype=np.int64)
        return cm.toarray()

    
class MiniBatchBinaryF1():
    def __init__(self):
        self.y_true = []
        self.y_pred = []
    
    def add_batch(self, batch_preds, batch_targets):
        """
        batch_preds: (N, F)
        batch_targets: (N,)
        """
        assert batch_preds.ndim == 2
        self.y_pred.append(batch_preds)
        self.y_true.append(batch_targets)
        
    def compute(self, pos_label=1):
        yp = torch.cat(self.y_pred).argmax(1)
        yt = torch.cat(self.y_true)
        
        neg_label = 0 if pos_label else 1        
        fn = ((yt == pos_label) & (yp == neg_label)).sum().item()
        tp = ((yt == pos_label) & (yp == pos_label)).sum().item()
        fp = ((yt == neg_label) & (yp == pos_label)).sum().item()
        return tp / (tp + (fp + fn)/2)


class MiniBatchBinaryPrecision():
    def __init__(self):
        self.y_true = []
        self.y_pred = []
    
    def add_batch(self, batch_preds, batch_targets):
        """
        batch_preds: (N, F)
        batch_targets: (N,)
        """
        assert batch_preds.ndim == 2
        self.y_pred.append(batch_preds)
        self.y_true.append(batch_targets)
        
    def compute(self, pos_label=1):
        yp = torch.cat(self.y_pred).argmax(1)
        yt = torch.cat(self.y_true)
        
        neg_label = 0 if pos_label else 1        
        tp = ((yt == pos_label) & (yp == pos_label)).sum().item()
        fp = ((yt == neg_label) & (yp == pos_label)).sum().item()
        return tp / (tp + fp)
    
    
class MiniBatchBinaryRecall():
    def __init__(self):
        self.y_true = []
        self.y_pred = []
    
    def add_batch(self, batch_preds, batch_targets):
        """
        batch_preds: (N, F)
        batch_targets: (N,)
        """
        assert batch_preds.ndim == 2
        self.y_pred.append(batch_preds)
        self.y_true.append(batch_targets)
        
    def compute(self, pos_label=1):
        yp = torch.cat(self.y_pred).argmax(1)
        yt = torch.cat(self.y_true)
        
        neg_label = 0 if pos_label else 1        
        fn = ((yt == pos_label) & (yp == neg_label)).sum().item()
        tp = ((yt == pos_label) & (yp == pos_label)).sum().item()
        return tp / (tp + fn)
    
    
class MiniBatchAccuracy():
    def __init__(self):
        self.y_true = []
        self.y_pred = []
    
    def add_batch(self, batch_preds, batch_targets):
        """
        batch_preds: (N, F)
        batch_targets: (N,)
        """
        assert batch_preds.ndim == 2
        self.y_pred.append(batch_preds)
        self.y_true.append(batch_targets)
        
    def compute(self):
        yp = torch.cat(self.y_pred).argmax(1)
        yt = torch.cat(self.y_true)
        return (yp == yt).sum().item() / yp.size(0)


class MiniBatchCost():
    def __init__(self):
        self.cost = 0
        self.total_data = 0

    def add_batch(self, batch_loss, n_data):
        """
        batch_preds: (N, F)
        batch_targets: (N,)
        """
        self.cost += batch_loss.item() * n_data
        self.total_data += n_data
        
    def compute(self):
        return self.cost / self.total_data


import torch
from torch.utils.data import Dataset

from jcopdl.exception import TruncationError


class TimeSeriesDataset(Dataset):
    """
    Time Series Dataset for PyTorch RNN

    == Arguments ==
    time_series: pandas DataFrame
        the time series data

    target_col_name: string
        target column name in the DataFrame

    seq_len: int
        sequence length. The length of input sequences for each row in the data

    target_lag: int
        how many lag to the future to predict.
        taget_lag = 1 means to predict the next data

    trunc_len: int
        sequence length for truncated BPTT.
        Note: seq_len should be multiple of trunc_len

    summary: bool
        print the batched dataset information
    """

    def __init__(self, time_series, target_col_name, seq_len, target_lag=1, trunc_len=None, summary=True):
        # n_step, n_seq, and n_truncate
        n_row, self.n_feature = time_series.shape
        self.n_seq = seq_len
        if trunc_len is None:
            self.truncated = False
            self.n_truncate = seq_len
        elif seq_len % trunc_len == 0:
            self.truncated = True
            self.n_truncate = trunc_len
        else:
            raise TruncationError("seq_len should be divisible by trunc_len for tBPTT to work")
        self.n_step = self.n_seq // self.n_truncate

        # Preparing time series for input and target
        input_series = time_series.iloc[:-target_lag, :]
        target_series = time_series[[target_col_name]].iloc[target_lag:, :]

        # Save index for plotting
        if hasattr(time_series, 'index'):
            self.input_ticks = input_series.index
            self.target_ticks = target_series.index

        # Preparing TNSF from series
        self.X = self._to_sequence(input_series)
        self.y = self._to_sequence(target_series)
        self.n_data = self.X.shape[0]

        # Convert to NSF if no truncation
        if not self.truncated:
            self.X.squeeze_(1)
            self.y.squeeze_(1)

        # Print Summary
        if summary:
            n_removed = n_row - target_lag - self.n_data * self.n_seq
            if self.truncated:
                print(f"(N, T, S, F): ({self.n_data}, {self.n_step}, {self.n_truncate}, {self.n_feature})")
            else:
                print(f"(N, S, F): ({self.n_data}, {self.n_seq}, {self.n_feature})")
            print(f"Note: last \x1b[31m{n_removed} data excluded\x1b[0m\n")

    def __getitem__(self, i):
        return self.X[i], self.y[i]

    def __len__(self):
        return len(self.X)

    def _to_sequence(self, series):
        dim = series.shape[1]
        series = torch.FloatTensor(series.values)
        series = series.unfold(0, self.n_seq, self.n_seq)
        series = series.view(-1, dim, self.n_step, self.n_truncate)
        series = series.permute(0, 2, 3, 1)
        return series



from tqdm.auto import tqdm
from jcopdl.metrics import MiniBatchConfusionMatrix, MiniBatchAccuracy
from jcopdl.visualization import visualize_prediction_batch

__all__ = [
    "evaluate_confusion_matrix",
    "evaluate_accuracy",
    "evaluate_prediction"
]


def evaluate_confusion_matrix(dataloader, model, device, desc=""):
    model.eval()
    metric = MiniBatchConfusionMatrix()
    for feature, target in tqdm(dataloader, desc=desc.title(), leave=False):
        feature, target = feature.to(device), target.to(device)
        output = model(feature)
        metric.add_batch(output, target)
    return metric.compute()


def evaluate_accuracy(dataloader, model, device, desc=""):
    model.eval()
    metric = MiniBatchAccuracy()
    for feature, target in tqdm(dataloader, desc=desc.title(), leave=False):
        feature, target = feature.to(device), target.to(device)
        output = model(feature)
        metric.add_batch(output, target)
    return metric.compute()


def evaluate_prediction(dataloader, model, device):
    model.eval()
    feature, target = next(iter(dataloader))
    feature, target = feature.to(device), target.to(device)
    output = model(feature)
    
    preds = output.argmax(1)
    classes = dataloader.dataset.classes
    image = visualize_prediction_batch(feature, target, preds, classes)
    return image



import json

snippet = {
    "name": "PyTorch x JCOpDL",
    "sub-menu": [{
            "name": "Import common packages",
            "snippet": [
                "import torch",
                "from torch import nn, optim",
                "from jcopdl.callback import Callback",
                "",
                "device = torch.device(\"cuda:0\" if torch.cuda.is_available() else \"cpu\")",
                "device"
            ]
        },
        {
            "name": "Dataset & Dataloader",
            "sub-menu": [{
                    "name": "Torchvision ImageFolder",
                    "snippet": [
                        "from torchvision import datasets, transforms",
                        "from torch.utils.data import DataLoader",
                        "",
                        "bs = \"______\"",
                        "crop_size = \"____\"",
                        "",
                        "train_transform = transforms.Compose([",
                        "    \"____________\",",
                        "    transforms.ToTensor()",
                        "])",
                        "",
                        "test_transform = transforms.Compose([",
                        "    \"____________\",",
                        "    transforms.ToTensor()",
                        "])",
                        "",
                        "train_set = datasets.ImageFolder(\"________\", transform=train_transform)",
                        "trainloader = DataLoader(train_set, batch_size=bs, shuffle=True)",
                        "",
                        "test_set = datasets.ImageFolder(\"________\", transform=test_transform)",
                        "testloader = DataLoader(test_set, batch_size=bs, shuffle=\"____\")",
                        "",
                        "",
                        "metadata = {",
                        "    \"crop_size\": crop_size",
                        "}"
                    ]
                },
                {
                    "name": "CharRNN Dataset",
                    "snippet": [
                        "from jcopdl import transforms",
                        "from jcopdl.utils.dataloader import CharRNNDataset, CharRNNDataloader",
                        "",
                        "train_set = CharRNNDataset(\"data / train.csv\", text_col=\"_______\", label_col=\"_______\", max_len=_______)",
                        "test_set = CharRNNDataset(\"data / test.csv\", text_col=\"_______\", label_col=\"_______\", chars=train_set.chars, classes=train_set.classes, pad=train_set.pad, max_len=_______)",
                        "",
                        "transform = transforms.Compose([",
                        "    transforms.PadSequence(),",
                        "    transforms.OneHotEncode(train_set.n_chars),",
                        "    transforms.TruncateSequence(200)",
                        "])",
                        "",
                        "trainloader = CharRNNDataloader(train_set, batch_size=16, batch_transform=transform, drop_last=True)",
                        "testloader = CharRNNDataloader(test_set, batch_size=16, batch_transform=transform, drop_last=True)",
                        "",
                        "",
                        "metadata = {",
                        "    \"chars\": train_set.chars,",
                        "    \"classes\": train_set.classes,",
                        "    \"pad\": train_set.pad",
                        "}"
                    ]
                }
            ]
        },
        {
            "name": "Arsitektur & Config",
            "sub-menu": [{
                    "name": "ANN Regression Example",
                    "snippet": [
                        "from torch import nn",
                        "from jcopdl.layers import linear_block",
                        "",
                        "class ANN(nn.Module):",
                        "    def __init__(self, input_size, n1, n2, output_size, dropout):",
                        "        super().__init__()",
                        "        self.fc = nn.Sequential(",
                        "            linear_block(input_size, n1, dropout=dropout),",
                        "            linear_block(n1, n2, dropout=dropout),",
                        "            linear_block(n2, output_size, activation=\"identity\")",
                        "        ),",
                        "    ",
                        "    def forward(self, x):",
                        "        return self.fc(x)",
                        "    ",
                        "",
                        "model_config = {",
                        "    \"input_size\": train_set.n_features,",
                        "    \"n1\": 128,",
                        "    \"n2\": 64,",
                        "    \"output_size\": 1,",
                        "    \"dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                },
                {
                    "name": "ANN Classification Example",
                    "snippet": [
                        "from torch import nn",                        
                        "from jcopdl.layers import linear_block",
                        "",
                        "class ANN(nn.Module):",
                        "    def __init__(self, input_size, n1, n2, output_size, dropout):",
                        "        super().__init__()",
                        "        self.fc = nn.Sequential(",
                        "            linear_block(input_size, n1, dropout=dropout),",
                        "            linear_block(n1, n2, dropout=dropout),",
                        "            linear_block(n2, output_size, activation=\"lsoftmax\")",
                        "        ),",
                        "    ",
                        "    def forward(self, x):",
                        "        return self.fc(x)",
                        "    ",
                        "",
                        "model_config = {",
                        "    \"input_size\": train_set.n_features,",
                        "    \"n1\": 128,",
                        "    \"n2\": 64,",
                        "    \"output_size\": 1,",
                        "    \"dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                },
                {
                    "name": "CNN Classification Example",
                    "snippet": [
                        "from torch import nn",                        
                        "from jcopdl.layers import linear_block, conv_block",
                        "",
                        "class CNN(nn.Module):",
                        "    def __init__(self, output_size, fc_dropout):",
                        "        super().__init__()",
                        "        self.conv = nn.Sequential(",
                        "            conv_block(\"___\", \"___\"),",
                        "            conv_block(\"___\", \"___\"),",
                        "            nn.Flatten()",
                        "        )",
                        "        ",
                        "        self.fc = nn.Sequential(",
                        "            linear_block(\"_____\", \"_____\", dropout=fc_dropout),",
                        "            linear_block(\"_____\", output_size, activation=\"lsoftmax\")",
                        "        )",
                        "        ",
                        "    def forward(self, x):",
                        "        return self.fc(self.conv(x))",
                        "    ",
                        "",
                        "model_config = {",
                        "    \"output_size\": len(train_set.classes),",
                        "    \"fc_dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                },
                {
                    "name": "Many to Many RNN Example",
                    "snippet": [
                        "from torch import nn",                        
                        "from jcopdl.layers import linear_block",
                        "",
                        "class ManytoManyRNN(nn.Module):",
                        "    def __init__(self, input_size, output_size, hidden_size, num_layers, dropout):",
                        "        super().__init__()",
                        "        self.rnn = nn.RNN(input_size, hidden_size, num_layers, dropout=dropout, batch_first=True)",
                        "        self.fc = linear_block(hidden_size, output_size, activation=\"identity\")",
                        "        ",
                        "    def forward(self, x, hidden):        ",
                        "        x, hidden = self.rnn(x, hidden)",
                        "        x = self.fc(x)",
                        "        return x, hidden",
                        "",
                        "",
                        "model_config = {",
                        "    \"input_size\": ________,",
                        "    \"output_size\": ________,",
                        "    \"hidden_size\": 64,",
                        "    \"num_layers\": 2,",
                        "    \"dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                },
                {
                    "name": "Many to One RNN Example",
                    "snippet": [
                        "from torch import nn",                        
                        "from jcopdl.layers import linear_block",
                        "",
                        "class ManyToOneRNN(nn.Module):",
                        "    def __init__(self, input_size, output_size, hidden_size, num_layers, dropout):",
                        "        super().__init__()",
                        "        self.rnn = nn.RNN(input_size, hidden_size, num_layers, dropout=dropout, batch_first=True)",
                        "        self.fc = linear_block(num_layers*hidden_size, output_size, activation=\"lsoftmax\")",
                        "        ",
                        "    def forward(self, x, hidden):        ",
                        "        x, hidden = self.rnn(x, hidden)",
                        "        n_layers, n_batch, n_hidden = hidden.shape",
                        "        last_state = hidden.permute(1, 0, 2).reshape(-1, n_layers*n_hidden) # LBH -> BLH -> BF",
                        "        x = self.fc(last_state)",
                        "        return x, hidden",
                        "",
                        "",
                        "model_config = {",
                        "    \"input_size\": ________,",
                        "    \"output_size\": ________,",
                        "    \"hidden_size\": 64,",
                        "    \"num_layers\": 2,",
                        "    \"dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                },
                {
                    "name": "Many to Many LSTM Example",
                    "snippet": [
                        "from torch import nn",                        
                        "from jcopdl.layers import linear_block",
                        "",
                        "class ManytoManyLSTM(nn.Module):",
                        "    def __init__(self, input_size, output_size, hidden_size, num_layers, dropout):",
                        "        super().__init__()",
                        "        self.rnn = nn.LSTM(input_size, hidden_size, num_layers, dropout=dropout, batch_first=True)",
                        "        self.fc = linear_block(hidden_size, output_size, activation=\"identity\")",
                        "        ",
                        "    def forward(self, x, hidden):        ",
                        "        x, hidden = self.rnn(x, hidden)",
                        "        x = self.fc(x)",
                        "        return x, hidden",
                        "",
                        "",
                        "model_config = {",
                        "    \"input_size\": ________,",
                        "    \"output_size\": ________,",
                        "    \"hidden_size\": 64,",
                        "    \"num_layers\": 2,",
                        "    \"dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                },
                {
                    "name": "Many to One LSTM Example",
                    "snippet": [
                        "from torch import nn",                        
                        "from jcopdl.layers import linear_block",
                        "",
                        "class ManyToOneLSTM(nn.Module):",
                        "    def __init__(self, input_size, output_size, hidden_size, num_layers, dropout):",
                        "        super().__init__()",
                        "        self.rnn = nn.LSTM(input_size, hidden_size, num_layers, dropout=dropout, batch_first=True)",
                        "        self.fc = linear_block(num_layers*2*hidden_size, output_size, activation=\"lsoftmax\")",
                        "        ",
                        "    def forward(self, x, hidden):        ",
                        "        x, (h, c) = self.rnn(x, hidden)",
                        "        state = torch.cat([h, c], dim=2)",
                        "        n_layers, n_batch, n_2hidden = state.shape",
                        "        last_state = state.permute(1, 0, 2).reshape(-1, n_layers*n_2hidden) # LBH -> BLH -> BF",
                        "        x = self.fc(last_state)",
                        "        return x, (h, c)",
                        "",
                        "",
                        "model_config = {",
                        "    \"input_size\": ________,",
                        "    \"output_size\": ________,",
                        "    \"hidden_size\": 64,",
                        "    \"num_layers\": 2,",
                        "    \"dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                },
                {
                    "name": "Many to Many GRU Example",
                    "snippet": [
                        "from torch import nn",                        
                        "from jcopdl.layers import linear_block",
                        "",
                        "class ManytoManyGRU(nn.Module):",
                        "    def __init__(self, input_size, output_size, hidden_size, num_layers, dropout):",
                        "        super().__init__()",
                        "        self.rnn = nn.GRU(input_size, hidden_size, num_layers, dropout=dropout, batch_first=True)",
                        "        self.fc = linear_block(hidden_size, output_size, activation=\"identity\")",
                        "        ",
                        "    def forward(self, x, hidden):        ",
                        "        x, hidden = self.rnn(x, hidden)",
                        "        x = self.fc(x)",
                        "        return x, hidden",
                        "",
                        "",
                        "model_config = {",
                        "    \"input_size\": ________,",
                        "    \"output_size\": ________,",
                        "    \"hidden_size\": 64,",
                        "    \"num_layers\": 2,",
                        "    \"dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                },
                {
                    "name": "Many to One GRU Example",
                    "snippet": [
                        "from torch import nn",                        
                        "from jcopdl.layers import linear_block",
                        "",
                        "class ManyToOneGRU(nn.Module):",
                        "    def __init__(self, input_size, output_size, hidden_size, num_layers, dropout):",
                        "        super().__init__()",
                        "        self.rnn = nn.GRU(input_size, hidden_size, num_layers, dropout=dropout, batch_first=True)",
                        "        self.fc = linear_block(num_layers*hidden_size, output_size, activation=\"lsoftmax\")",
                        "        ",
                        "    def forward(self, x, hidden):        ",
                        "        x, hidden = self.rnn(x, hidden)",
                        "        n_layers, n_batch, n_hidden = hidden.shape",
                        "        last_state = hidden.permute(1, 0, 2).reshape(-1, n_layers*n_hidden) # LBH -> BLH -> BF",
                        "        x = self.fc(last_state)",
                        "        return x, hidden",
                        "",
                        "",
                        "model_config = {",
                        "    \"input_size\": ________,",
                        "    \"output_size\": ________,",
                        "    \"hidden_size\": 64,",
                        "    \"num_layers\": 2,",
                        "    \"dropout\": 0",
                        "}",
                        "optimizer_config = {",
                        "    \"lr\": 0.001",
                        "}"
                    ]
                }
            ]
        },
        {
            "name": "Training Preparation (MCOC)",
            "snippet": [
                "model = _______(**model_config).to(device)",
                "criterion = _______",
                "optimizer = optim.AdamW(model.parameters(), **optimizer_config)",
                "callback = Callback(model, optimizer, model_config, optimizer_config, metadata, outdir=\"output\")"
            ]
        },
        {
            "name": "Add Plot to Callback",
            "sub-menu": [{
                    "name": "Cost",
                    "snippet": [
                        "callback.add_plot([\"train_cost\", \"test_cost\"], scale=\"semilogy\")"
                    ]
                },
                {
                    "name": "Score",
                    "snippet": [
                        "callback.add_plot([\"train_score\", \"test_score\"], scale=\"linear\")"
                    ]
                },
                {
                    "name": "Image",
                    "snippet": [
                        "callback.add_image(\"test_predict\")"
                    ]
                }
            ]
        },
        {
            "name": "Training Loop Function",
            "sub-menu": [{
                    "name": "Basic + Cost + Acc",
                    "snippet": [
                        "from tqdm.auto import tqdm",
                        "from jcopdl.metrics import MiniBatchCost, MiniBatchAccuracy",
                        "from jcopdl.visualization import visualize_prediction_batch",
                        "",
                        "",
                        "def train_loop_fn(mode, dataloader, model, criterion, optimizer, callback, device):",
                        "    if mode == \"train\":",
                        "        model.train()",
                        "    elif mode == \"test\":",
                        "        model.eval()",
                        "    ",
                        "    cost = MiniBatchCost()",
                        "    score = MiniBatchAccuracy()",
                        "    for feature, target in tqdm(dataloader, desc=mode.title(), leave=False):",
                        "        feature, target = feature.to(device), target.to(device)",
                        "        output = model(feature)",
                        "        loss = criterion(output, target)",
                        "        ",
                        "        if mode == \"train\":",
                        "            loss.backward()",
                        "            optimizer.step()",
                        "            optimizer.zero_grad()",
                        "",
                        "        cost.add_batch(loss, feature.size(0))",
                        "        score.add_batch(output, target)",
                        "    callback.log(f\"{mode}_cost\", cost.compute())",
                        "    callback.log(f\"{mode}_score\", score.compute())",
                        "    ",
                        "    if mode == \"test\":",
                        "        preds = output.argmax(1)",
                        "        classes = dataloader.dataset.classes",
                        "        image = visualize_prediction_batch(feature, target, preds, classes)",
                        "        callback.log_image(\"test_predict\", image)"
                    ]
                },
                {
                    "name": "Basic + Cost + F1",
                    "snippet": [
                        "from tqdm.auto import tqdm",
                        "from jcopdl.metrics import MiniBatchCost, MiniBatchBinaryF1",
                        "",
                        "",
                        "def train_loop_fn(mode, dataloader, model, criterion, optimizer, callback, device):",
                        "    if mode == \"train\":",
                        "        model.train()",
                        "    elif mode == \"test\":",
                        "        model.eval()",
                        "    ",
                        "    cost = MiniBatchCost()",
                        "    score = MiniBatchBinaryF1()",
                        "    for feature, target in tqdm(dataloader, desc=mode.title(), leave=False):",
                        "        feature, target = feature.to(device), target.to(device)",
                        "        output = model(feature)",
                        "        loss = criterion(output, target)",
                        "        ",
                        "        if mode == \"train\":",
                        "            loss.backward()",
                        "            optimizer.step()",
                        "            optimizer.zero_grad()",
                        "",
                        "        cost.add_batch(loss, feature.size(0))",
                        "        score.add_batch(output, target)",
                        "    callback.log(f\"{mode}_cost\", cost.compute())",
                        "    callback.log(f\"{mode}_score\", score.compute(pos_label=1))"                   
                    ]
                },
                {
                    "name": "RNN + Cost + F1",
                    "snippet": [
                        "from tqdm.auto import tqdm",
                        "from torch.nn.utils import clip_grad_norm_",
                        "from jcopdl.metrics import MiniBatchCost, MiniBatchBinaryF1",
                        "",
                        "",
                        "def train_loop_fn(mode, dataloader, model, criterion, optimizer, callback, device):",
                        "    if mode == \"train\":",
                        "        model.train()",
                        "    elif mode == \"test\":",
                        "        model.eval()",
                        "    ",
                        "    cost = MiniBatchCost()",
                        "    score = MiniBatchBinaryF1()",
                        "    for feature, target in tqdm(dataloader, desc=mode.title(), leave=False):",
                        "        feature, target = feature.to(device), target.to(device)",
                        "        output, hidden = model(feature, None)",
                        "        loss = criterion(output, target)",
                        "        ",
                        "        if mode == \"train\":",
                        "            loss.backward()",
                        "            clip_grad_norm_(model.parameters(), 2)",
                        "            optimizer.step()",
                        "            optimizer.zero_grad()",
                        "",
                        "        cost.add_batch(loss, feature.size(0))",
                        "        score.add_batch(output, target)",
                        "    callback.log(f\"{mode}_cost\", cost.compute())",
                        "    callback.log(f\"{mode}_score\", score.compute(pos_label=1))"
                    ]
                },
                {
                    "name": "RNN + TBPTT + Cost + F1",
                    "snippet": [
                        "from tqdm.auto import tqdm",
                        "from torch.nn.utils import clip_grad_norm_",
                        "from jcopdl.metrics import MiniBatchCost, MiniBatchBinaryF1",
                        "",
                        "",
                        "def train_loop_fn(mode, dataloader, model, criterion, optimizer, callback, device):",
                        "    if mode == \"train\":",
                        "        model.train()",
                        "    elif mode == \"test\":",
                        "        model.eval()",
                        "    ",
                        "    cost = MiniBatchCost()",
                        "    score = MiniBatchBinaryF1()",
                        "    for (prior, feature), target in tqdm(dataloader, desc=mode.title(), leave=False):",
                        "        prior, feature, target = prior.to(device), feature.to(device), target.to(device)",
                        "        with torch.no_grad():",
                        "            output, hidden = model(prior, None)",
                        "        output, hidden = model(feature, hidden)",
                        "        loss = criterion(output, target)",
                        "        ",
                        "        if mode == \"train\":",
                        "            loss.backward()",
                        "            clip_grad_norm_(model.parameters(), 2)",
                        "            optimizer.step()",
                        "            optimizer.zero_grad()",
                        "",
                        "        cost.add_batch(loss, feature.size(0))",
                        "        score.add_batch(output, target)",
                        "    callback.log(f\"{mode}_cost\", cost.compute())",
                        "    callback.log(f\"{mode}_score\", score.compute(pos_label=1))"
                    ]
                }
            ]
        },
        {
            "name": "Training",
            "sub-menu": [{
                    "name": "Minimize Cost",
                    "snippet": [
                        "while True:",
                        "    train_loop_fn(\"train\", trainloader, model, criterion, optimizer, callback, device)",
                        "    with torch.no_grad():",
                        "        train_loop_fn(\"test\", testloader, model, criterion, optimizer, callback, device)",
                        "    ",
                        "    if callback.early_stopping(\"minimize\", \"test_cost\"):",
                        "        break"
                    ]
                },
                {
                    "name": "Maximize Score",
                    "snippet": [
                        "while True:",
                        "    train_loop_fn(\"train\", trainloader, model, criterion, optimizer, callback, device)",
                        "    with torch.no_grad():",
                        "        train_loop_fn(\"test\", testloader, model, criterion, optimizer, callback, device)",
                        "    ",
                        "    if callback.early_stopping(\"maximize\", \"test_score\"):",
                        "        break"
                    ]
                }
            ]
        },
        {
            "name": "Evaluation",
            "sub-menu": [{
                    "name": "Visualize Prediction",
                    "snippet": [
                        "from jcopdl.eval import evaluate_prediction",
                        "",                       
                        "with torch.no_grad():",
                        "    img_train = evaluate_prediction(trainloader, model, device)",
                        "    img_test = evaluate_prediction(testloader, model, device)"
                    ]
                },{
                    "name": "Evaluate Accuracy",
                    "snippet": [
                        "from jcopdl.eval import evaluate_accuracy",
                        "",                             
                        "with torch.no_grad():",
                        "    acc_train = evaluate_accuracy(trainloader, model, device)",
                        "    acc_test = evaluate_accuracy(testloader, model, device)"
                    ]
                },{
                    "name": "Evaluate Confusion Matrix",
                    "snippet": [
                        "from jcopdl.eval import evaluate_confusion_matrix",
                        "from jcopdl.visualization import plot_confusion_matrix",
                        "",
                        "with torch.no_grad():",
                        "    cm_train = evaluate_confusion_matrix(trainloader, model, device)",
                        "    cm_test = evaluate_confusion_matrix(testloader, model, device)",
                        "",
                        "fig_train = plot_confusion_matrix(cm_train, train_set.classes)",
                        "fig_test = plot_confusion_matrix(cm_test, test_set.classes)"
                    ]
                }
            ]
        },
        {
            "name": "Load Model",
            "sub-menu": [{
                    "name": "Continue from Checkpoint",
                    "snippet": [
                        "from jcopdl.io import load_from_checkpoint",
                        "",                       
                        "model, optimizer, callback = load_from_checkpoint(\"output/checkpoint/_________.pth\")",
                        "criterion = _______"
                    ]
                },{
                    "name": "Load best model",
                    "snippet": [
                        "import torch",
                        "device = torch.device(\"cuda:0\" if torch.cuda.is_available() else \"cpu\")",
                        "",                             
                        "model = torch.load(\"output/model_best.pth\", map_location=\"cpu\").to(device)"
                    ]
                },{
                    "name": "Load config, metadata, etc",
                    "snippet": [
                        "import torch",
                        "",
                        "model_config = torch.load(\"output/model_configs.pth\")",
                        "optimizer_config = torch.load(\"output/optimizer_configs.pth\")",
                        "metadata = torch.load(\"output/metadata.pth\")"
                    ]
                }
            ]
        }
    ]
}

print(json.dumps(snippet, indent=4))



import io
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
import seaborn as sns

import torch
from torchvision.utils import make_grid
from torchvision.transforms.functional import to_pil_image


def visualize_image_batch(images, n_col=8):
    if isinstance(images, list):
        images = torch.stack(images, dim=0)    
    grid = make_grid(images, nrow=n_col)
    return to_pil_image(grid)


def visualize_prediction_batch(images, labels, preds, classes=None, image_scale=2, fontsize=8, bg_color="white", to_pillow=True):
    n_data = images.size(0)
    assert len(labels) == len(preds) == n_data
    preds, labels = preds.cpu(), labels.cpu()

    if classes is not None:
        classes = np.array(classes)
        preds = classes[preds]
        labels = classes[labels]

    # compute n_row, n_col and figsize
    n_col = 8
    n_row = n_data // n_col + (n_data % n_col > 0)
    figsize=(image_scale*n_col, image_scale*n_row)

    fig, axes = plt.subplots(n_row, n_col, figsize=figsize)
    fig.set_facecolor(bg_color)
    fig.subplots_adjust(wspace=0.05, hspace=0.15)
    for ax in axes.flatten():
        ax.axis('off');

    for image, label, pred, ax in zip(images, labels, preds, axes.flatten()):
        ax.imshow(image.permute(1, 2, 0).cpu())
        font = {"color": 'r', "fontsize": fontsize} if label != pred else {"color": 'g', "fontsize": fontsize}
        ax.set_title(f"{label} | P: {pred}", fontdict=font);
    fig.tight_layout()

    if to_pillow:
        img_buf = io.BytesIO()
        fig.savefig(img_buf, format='png');
        plt.close()
        fig = Image.open(img_buf)
    return fig


def plot_confusion_matrix(cm, labels, figsize=(8, 8), fontsize=10, colormap="Blues"):
    acc = cm.trace() / cm.sum()
    
    fig = plt.figure(figsize=figsize)
    sns.heatmap(cm, annot=True, square=True, cmap=colormap, cbar=False, xticklabels=labels, yticklabels=labels,
                fmt="d", annot_kws={"fontsize": fontsize})
    plt.title(f'Accuracy: {acc:.3f}', fontsize=fontsize+1)
    plt.xlabel('Prediction', fontsize=fontsize+1)
    plt.ylabel('Actual', fontsize=fontsize+1)
    plt.yticks(rotation=0, verticalalignment='center')
    return fig



from jcopdl.layers._block import LinearBlock, ConvBlock, TConvBlock


def linear_block(n_in, n_out, activation='relu', batch_norm=False, dropout=0.):
    return LinearBlock(n_in, n_out, activation, batch_norm, dropout)


def conv_block(c_in, c_out, kernel=3, stride=1, pad=1, bias=True, activation='relu', batch_norm=False, dropout=0,
               pool_type='maxpool', pool_kernel=2, pool_stride=2):
    return ConvBlock(c_in, c_out, kernel, stride, pad, bias, activation, batch_norm, dropout, pool_type, pool_kernel,
                     pool_stride)


def tconv_block(c_in, c_out, kernel=4, stride=2, pad=1, bias=True, activation='relu', batch_norm=False, dropout=0,
                pool_type=None, pool_kernel=2, pool_stride=2):
    return TConvBlock(c_in, c_out, kernel, stride, pad, bias, activation, batch_norm, dropout, pool_type, pool_kernel,
                      pool_stride)



import torch


def load_from_checkpoint(checkpoint_path):
    callback = torch.load(checkpoint_path)
    callback.epoch += 1
    model = callback.model
    optimizer = callback.optimizer
    return model, optimizer, callback



def in_colab():
    "Check if the code is running in Google Colaboratory"
    try:
        from google import colab
        return True
    except: 
        return False


def in_notebook():
    "Check if the code is running in a jupyter notebook"
    if in_colab():
        return True
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell': # Jupyter notebook, Spyder or qtconsole
            import IPython
            # IPython version lower then 6.0.0 don't work with output you update
            return IPython.__version__ >= '6.0.0'
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter

IN_NOTEBOOK = in_notebook()



import torch
import numpy as np
import pandas as pd
from copy import deepcopy
from PIL import Image
from pathlib import Path
from warnings import warn

from jcopdl.utils.nb_check import IN_NOTEBOOK
from jcopdl.utils.helper import listify
from jcopdl.exception import NotLoggedError, InvalidScaleValue, InvalidSenseValue, ModelError

if IN_NOTEBOOK:
    try:
        from IPython.display import display, HTML
        import matplotlib.pyplot as plt
    except:
        warn("Failed to import ipywidgets, fallback to console behavior")
        IN_NOTEBOOK = False


class Callback:
    """
    Callback for common PyTorch Workflow:
    - Neat Checkpoint and Logs
    - Early stopping
    - Runtime Plotting
    - Runtime Log and Reporting


    == Arguments ==
    model: torch.nn.Module
        A deep learning architecture using PyTorch nn.Module

    model_config: dict
        A dictionary with model parameters. It would be saved in callback.

    metadata: dict
        Any other value you would want to save in callback. Usually training and dataloader parameters.

    checkpoint_every: int
        Number of epoch to save a checkpoint. Checkpoint would be deleted after the training finished.

    early_stop_patience: int
        number of patience tolerance before executing early stopping

    max_epoch: int
        Limit the training to this epoch. Use None if you want to train indefinitely.

    outdir: string
        path of output directory to save the weights, configs, and logs. You may also include subpath such as `outdir/model1` or `outdir/model2`


    == Example Usage ==
    # Initialization
    callback = Callback(model, config, outdir="model")

    # Add single metric plot
    callback.add_plot("test_cost", scale="semilogy")

    # Add multi-metric in single plot
    callback.add_plot(["train_cost", "test_cost"], scale="semilogy")

    # Add multi-metric in multi subplot
    callback.add_plot(["train_cost", "test_cost"], scale="semilogy")
    callback.add_plot(["train_acc", "test_acc"], scale="linear")

    # Add image log
    callback.add_image("test_image")

    # Logging
    callback.log("test_cost", value)

    # Early Stopping
    if callback.early_stopping("test_cost"):
        break
    """
    def __init__(self, model, optimizer, model_config=None, optimizer_config=None, metadata=None, checkpoint_every=50, early_stop_patience=5, max_epoch=None, outdir="output"):
        self.outdir = Path(outdir)   
        self.outdir.mkdir(parents=True, exist_ok=True)
        
        self._check_value(model, model_config, optimizer_config, metadata, max_epoch)
        
        self.model = model
        self.optimizer = optimizer
        self.model_config = model_config
        self.optimizer_config = optimizer_config
        self.metadata = metadata
        
        self.checkpoint_every = checkpoint_every
        self.early_stop_patience = early_stop_patience
        self.best_metric = None
        self.best_epoch = None

        self.logs = {}
        self.epoch = 1
        self.early_stop = 0
        
        self._report_nb = None

        self._graph_nb = None
        self._graph_fig = None
        self._graph_axes = None        
        self._graph_metrics = []
        self._graph_scales = []

        self._image_nb = None
        self._image_metrics = []

    def early_stopping(self, sense, monitor, load_best_when_stop=True, plot=True, plot_image=True, figsize="auto"):
        """
        sense: {"maximize", "minimize"}
            Sense of the metric's.
            - use "minimize" if lower is better
            - use "maximize" if higher is better

        monitor: str
            metric benchmark for early stopping        
        """
        if not IN_NOTEBOOK:
            plot = False
            plot_image = False
        
        if monitor not in self.logs:
            raise NotLoggedError(f'Metric {monitor} is not logged. Please use callback.log("{monitor}", value).')

        # Show report
        self.log("epoch", self.epoch)
        if IN_NOTEBOOK:
            report_table = HTML(pd.DataFrame(self.logs).set_index("epoch").to_html())
            if self._report_nb is None:
                self._report_nb = display(report_table, display_id=True)
            else:
                self._report_nb.update(report_table)
        else:
            report = f'Epoch {self.epoch:5}\n'
            report += " | ".join([f'{k.title():10} = {v[-1]:.4f}' for k, v in self.logs.items() if k != "epoch"])
            print(report + "\n")        
            
        # Save checkpoint
        if self.epoch % self.checkpoint_every == 0:
            self._save_ckpt()
            
        # Early stopping logic
        if sense not in ["maximize", "minimize"]:
            raise InvalidSenseValue('Available sense {"maximize", "minimize"}')
        
        if self.best_metric is None:
            self.best_metric = 0 if sense == "maximize" else np.inf
            self.best_epoch = 1

        last_metric = self.logs[monitor][-1]
        is_improve = last_metric > self.best_metric if sense == "maximize" else last_metric < self.best_metric
            
        stop = False
        if is_improve:
            self.best_metric = last_metric
            self.best_epoch = self.epoch
            self.reset_early_stop()
            self._save_best()
        else:
            self.early_stop += 1
            if not IN_NOTEBOOK:
                print(f"\x1b[31m==> EarlyStop patience = {self.early_stop:2} | Best {monitor}: {self.best_metric:.4f}\x1b[0m")

            if self.early_stop == self.max_epoch:
                print(f'\x1b[31m==> Max epoch reached at epoch: {self.epoch} | Best {monitor}: {self.best_metric:.4f}\x1b[0m')
                print(f'\x1b[31m==> Best model is saved at {self.outdir}\x1b[0m')
                stop = True                
            elif self.early_stop >= self.early_stop_patience:
                print(f'\x1b[31m==> Execute Early Stopping at epoch: {self.epoch} | Best {monitor}: {self.best_metric:.4f}\x1b[0m')
                print(f'\x1b[31m==> Best model is saved at {self.outdir}\x1b[0m')
                stop = True

        # Plot if there are any registered metrics
        if plot:
            with plt.style.context("seaborn-bright"):
                self._plot_metrics(monitor, figsize)

        if plot_image:
            self._plot_images()
                
        # Cleanup and load best when stop                
        if stop and load_best_when_stop:
            self.model = torch.load(self.outdir / "model_best.pth", map_location="cpu")
            self.optimizer = torch.load(self.outdir / "optimizer_best.pth", map_location="cpu")
                
        self.epoch += 1
        return stop

    def add_plot(self, metrics, scale="linear"):
        """
        Please make sure that the registered metrics is logged.
        Available scale: {"linear", "semilogy"}
        """
        if scale not in ["linear", "semilogy"]:
            raise InvalidScaleValue('Available scale {"linear", "semilogy"}')
            
        metrics = listify(metrics)
        self._graph_metrics.append(metrics)
        self._graph_scales.append(scale)

    def add_image(self, metrics): 
        metrics = listify(metrics)
        for metric in metrics:
            log_path = self.outdir / "image_logs" / metric
            log_path.mkdir(parents=True, exist_ok=True)
        self._image_metrics.extend(metrics)
        
    def _plot_metrics(self, monitor, figsize):
        # Skip plot if no registered metrics
        if not bool(self._graph_metrics):
            return

        if figsize == "auto":
            figsize = (7*len(self._graph_metrics), 5)
        
        if self._graph_nb is None:
            self._graph_fig, axes = plt.subplots(1, len(self._graph_metrics), figsize=figsize)
            if len(self._graph_metrics) == 1:
                axes = [axes] 
            self._graph_axes = axes

            for ax, metrics, scale in zip(self._graph_axes, self._graph_metrics, self._graph_scales):
                ax.set_xlabel("Epoch")
                for metric in metrics:
                    self._plot(ax, self.logs["epoch"], self.logs[metric], scale, metric)
                ax.legend()                        
            self._graph_nb = display(self._graph_fig, display_id=True)
        else:
            for ax, metrics, scale in zip(self._graph_axes, self._graph_metrics, self._graph_scales):
                ax.clear()
                ax.set_xlabel("Epoch")  
                for metric in metrics:
                    self._plot(ax, self.logs["epoch"], self.logs[metric], scale, metric)
                    if metric == monitor:
                        self._plot(ax, self.best_epoch, self.best_metric, "scatter", "best")
                        ax.set_title(f"Patience: {self.early_stop:2} | Best {monitor}: {self.best_metric:.4f}", fontsize=12)
                ax.legend()
            self._graph_nb.update(self._graph_fig)

    def _plot_images(self):
        # Skip plot if no registered metrics
        if not bool(self._image_metrics):
            return
        
        if self._image_nb is None:
            self._image_nb = [display(Image.open(self.logs[metric][-1]), display_id=True) for metric in self._image_metrics]
        else:
            for nb, metric in zip(self._image_nb, self._image_metrics):
                nb.update(Image.open(self.logs[metric][-1]))
                
    def log(self, name, value):
        if type(value) == torch.Tensor:
            value = value.item()
        if name not in self.logs:
            self.logs[name] = []
        self.logs[name].append(value)

    def log_image(self, name, pil_image):
        if name not in self.logs:
            self.logs[name] = []
        log_path = self.outdir / "image_logs" / name / f"{self.epoch}.png"
        pil_image.save(log_path)
        self.logs[name].append(log_path)

    def reset_early_stop(self):
        self.early_stop = 0
        
    def change_early_stop_patience(self, patience):
        self.early_stop_patience = patience

    def _save_best(self):
        torch.save(self.model, self.outdir / f'model_best.pth')
        torch.save(self.optimizer, self.outdir / f'optimizer_best.pth')
        try:
            pd.DataFrame(self.logs).to_csv(self.outdir / "logs.csv", index=None)
        except:
            torch.save(self.logs, self.outdir / "logs.pth")

    def _save_ckpt(self):
        ckpt_path = self.outdir / "checkpoint"
        ckpt_path.mkdir(parents=True, exist_ok=True)
        
        callback = deepcopy(self)
        callback._report_nb = None
        callback._graph_nb = None
        callback._image_nb = None
        callback._graph_fig = None
        callback._graph_axes = None
        torch.save(callback, ckpt_path / f'epoch_{self.epoch}.pth')
        del(callback)

    def _check_value(self, model, model_config, optimizer_config, metadata, max_epoch):
        self.max_epoch = max_epoch if max_epoch is not None else np.inf
        
        if (model_config is not None) and isinstance(model_config, dict):
            torch.save(model_config, self.outdir / "model_configs.pth")

        if (optimizer_config is not None) and isinstance(optimizer_config, dict):
            torch.save(optimizer_config, self.outdir / "optimizer_configs.pth")            
            
        if (metadata is not None) and isinstance(metadata, dict):
            torch.save(metadata, self.outdir / "metadata.pth")            

        if model.__module__ == "__main__":            
            raise ModelError(f'Please write model in script (ex: model.py) and import it properly.\n>> from model import {model._get_name()}')

    @staticmethod
    def _plot(ax, x, y, scale, label):
        if scale == "linear":
            ax.plot(x, y, label=label)
        elif scale == "semilogy":
            ax.semilogy(x, y, label=label)
        elif scale == "scatter":
            ax.scatter(x, y, label=label)


from torch import nn


class BaseBlock(nn.Module):
    def __init__(self):
        super().__init__()
        self.block = None

        self._activation_map = {
            "relu": nn.ReLU(),
            "lrelu": nn.LeakyReLU(),
            "sigmoid": nn.Sigmoid(),
            "tanh": nn.Tanh(),
            "elu": nn.ELU(),
            "selu": nn.SELU(),
            "lsoftmax": nn.LogSoftmax(dim=1),
            "identity": nn.Identity(),
            "softmax": nn.Softmax(dim=1)
        } 

    def append_activation(self, activation):
        if isinstance(activation, str):
            if activation not in self._activation_map:
                raise Exception(f"jcopdl supports these activations ({', '.join(self._activation_map.keys())}, None)")
            self.block.add_module(activation.title(), self._activation_map[activation])
        elif activation.__module__ == "torch.nn.modules.activation":
            self.block.add_module(activation._get_name(), activation)

    def append_dropout(self, dropout):
        self.block.add_module("do", nn.Dropout(dropout))

    def append_dropout2d(self, dropout):
        self.block.add_module("do", nn.Dropout2d(dropout))

    def append_batchnorm(self, n):
        self.block.add_module("bn", nn.BatchNorm1d(n))

    def append_batchnorm2d(self, n):
        self.block.add_module("bn", nn.BatchNorm2d(n))

    def append_pooling(self, pool, pool_kernel, pool_stride):
        if pool == "maxpool":
            self.block.add_module("maxpool", nn.MaxPool1d(pool_kernel, pool_stride))
        elif pool == "avgpool":
            self.block.add_module("avgpool", nn.AvgPool1d(pool_kernel, pool_stride))
        elif pool is not None:
            raise Exception("jcopdl supports these pooling ({maxpool, avgpool, None})")

    def append_pooling2d(self, pool, pool_kernel, pool_stride):
        if pool == "maxpool":
            self.block.add_module("maxpool", nn.MaxPool2d(pool_kernel, pool_stride))
        elif pool == "avgpool":
            self.block.add_module("avgpool", nn.AvgPool2d(pool_kernel, pool_stride))
        elif pool is not None:
            raise Exception("jcopdl supports these pooling ({maxpool, avgpool, None})")



import torch


class Standardize(torch.nn.Module):
    """Standardize image using Standard Scaling."""
    def __init__(self, normalize=True):
        super().__init__()
        self.normalize = normalize

    def forward(self, x):
        bs, c = x.shape[:2]
        tmp = x.view(bs, c, -1)
        mean, std = tmp.mean(2, keepdims=True), tmp.std(2, keepdims=True)
        x = (x - mean) / (std + 1e-6)

        if self.normalize:
            x = (x + 5) / 10
            x = x.clamp_(0, 1)
        return x



import os
import random
import torch
import numpy as np


def set_pytorch_random_state(seed=42):
    print(f"Random state: {seed}")
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True


def pytorch_state(seed, device):
    return torch.Generator(device).manual_seed(seed)



from .augment import Standardize
from .sequence import OneHotEncode, TruncateSequence, PadSequence


class Compose:
    """
    Composes several transforms together. 

    == Arguments ==
    transform = transforms.Compose([
        transforms.PadSequence(),
        transforms.OneHotEncode(5),
        transforms.TruncateSequence(200)
    ])
    """
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, x):
        for t in self.transforms:
            x = t(x)
        return x

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string


class TruncationError(Exception):
    pass

class NotLoggedError(Exception):
    pass

class InvalidSenseValue(Exception):
    pass

class InvalidScaleValue(Exception):
    pass

class ModelError(Exception):
    pass



def conv_output_size(input_size, kernel, stride, pad):
    return ((input_size - kernel + 2 * pad) // stride) + 1


def tconv_output_size(input_size, kernel, stride, pad):
    return (input_size - 1) * stride + kernel - 2 * pad



import os

import PIL
import pandas as pd
import torch
from torch.utils.data import Dataset


class MultilabelDataset(Dataset):
    """
    Multilabel Dataset for PyTorch
    The folder structur should be
        _________________________________
        data/
            train/
                img1.jpg
                2.jpg
                three.jpg
                ... all train images ...
            test/
                a.jpg
                b.jpg
                c.jpg
                ... all test images ...
            train_label.csv
            test_label.csv
        _________________________________

    train_label.csv and test_label.csv are the metadata. It should contain file name and its corresponding labels
    For example:
        __________________________
        fname,label1,label2,label3
        img1.jpg,1,1,0
        2.jpg,1,0,0
        three.jpg,1,1,1
        ...
        __________________________

    == Arguments ==
    csv_path: string
        the metadata file

    csv_path: string
        the img folder

    transform: torchvision.transform
        torchvision data augmentation

    fname_col: string
        header of file name used in the metadata
    """
    def __init__(self, csv_path, img_path, transform=None, fname_col='fname'):
        df = pd.read_csv(csv_path)
        assert df[fname_col].apply(lambda x: os.path.isfile(img_path + x)).all

        self.classes = df.columns[1:].to_list()

        self.img_path = img_path
        self.transform = transform

        self.X = df[fname_col]
        self.y = df.drop(columns=fname_col)

    def __getitem__(self, index):
        img = PIL.Image.open(self.img_path + self.X[index]).convert("RGB")
        if self.transform is not None:
            img = self.transform(img)

        label = torch.from_numpy(self.y.iloc[index].values).type(torch.FloatTensor)
        return img, label

    def __len__(self):
        return len(self.X)



import numpy as np

import torch
from torch.utils.data import DataLoader
from torchvision.datasets.coco import CocoDetection
import matplotlib.pyplot as plt


class CocoDetectionDataset(CocoDetection):
    def __init__(self, root, annFile, device, transforms=None, bbox_format="voc"):
        super().__init__(root, annFile)
        self.transforms = transforms
        self.bbox_format = bbox_format
        self.classes_to_idx = {"background": 0}
        for item in self.coco.cats.values():
            self.classes_to_idx[item["name"]] = item["id"] + 1
        self.classes = list(self.classes_to_idx)
        self.device = device

    def __getitem__(self, index):
        image_id = self.ids[index]
        image = self._get_image(image_id)
        annots = self._get_annots(image_id)

        annots["boxes"] = [self._clip_bbox(box, image) for box in annots["boxes"]]

        if self.transforms is not None:
            sample = self.transforms(image=image, bboxes=annots["boxes"], labels=annots["labels"])
            image = sample["image"]
            annots["boxes"] = sample["bboxes"]
            annots["labels"] = sample["labels"]

        if self.bbox_format == "voc":
            annots["boxes"] = [self._coco_to_voc(box) for box in annots["boxes"]]

        image = torch.FloatTensor(image / 255).permute(2, 0, 1)
        annots["boxes"] = torch.FloatTensor(annots["boxes"]).to(self.device)
        annots["labels"] = torch.LongTensor(annots["labels"]).to(self.device) + 1
        return image, annots

    def _get_image(self, image_id):
        path = self.coco.loadImgs(image_id)[0]['file_name']
        image = plt.imread(f"{self.root}/{path}")
        return image

    def _get_annots(self, image_id):
        annot_ids = self.coco.getAnnIds(imgIds=image_id)
        annots = self.coco.loadAnns(annot_ids)
        return {
            "id": annot_ids,
            "image_id": image_id,
            "labels": [ann["category_id"] for ann in annots],
            "segmentation": [ann["segmentation"] for ann in annots],
            "boxes": [ann["bbox"] for ann in annots],
            "ignore": [ann["ignore"] for ann in annots],
            "iscrowd": [ann["iscrowd"] for ann in annots],
            "area": [ann["area"] for ann in annots]
        }

    @staticmethod
    def _coco_to_voc(bbox):
        x1, y1, w, h = bbox
        return x1, y1, x1 + w, y1 + h

    @staticmethod
    def _clip_bbox(bbox, image):
        h, w, c = image.shape
        x1, y1, _, _ = bbox
        bbox = np.clip(bbox, [0, 0, 0, 0], [w, h, w-x1, h-y1])
        return bbox


class CocoDetectionDataloader(DataLoader):
    def __init__(self, dataset, batch_size=1, shuffle=False, drop_last=False):
        super().__init__(dataset, batch_size, shuffle, collate_fn=self.collate, num_workers=0, drop_last=drop_last)
        self.device = dataset.device

    def collate(self, batch):
        images, annots = zip(*batch)
        images = torch.stack(list(images)).to(self.device)
        return images, list(annots)



from .time_series import TimeSeriesDataset
from .multilabel import MultilabelDataset
from .coco_detection import CocoDetectionDataset, CocoDetectionDataloader
from .char_rnn import CharRNNDataset, CharRNNDataloader
